/*
 * beatfish.c
 *
 * Copyright (C) 2003 Conrad Parker <conrad@metadecks.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 *
 */  

#include "config.h"
#include "beatfish.h"

#include <pthread.h>
#include <remix/remix.h>
#include <jack/jack.h>

/* for getpid() */
#include <sys/types.h>
#include <unistd.h>

#define IM PACKAGE_DATA_DIR "/"

#define USE_JACK

#define TINY_CURSOR

#define WIDTH 640
#define HEIGHT 480 

#define SAMPLES_DIR PACKAGE_DATA_DIR "/"

#define KICK_PATH SAMPLES_DIR "housekick1.wav"
#define SNARE_PATH SAMPLES_DIR "crazysnare1.wav"
#define CONGAHIGH_PATH SAMPLES_DIR "congahigh1.wav"
#define CONGALOW_PATH SAMPLES_DIR "congalow1.wav"
#define CLOSEDHAT_PATH SAMPLES_DIR "closedhat10.wav"
#define OPENHAT_PATH SAMPLES_DIR "hr16openhat.wav"
#define CLAP_PATH SAMPLES_DIR "clap45.wav"
#define CLAVE_PATH SAMPLES_DIR "hr16clave.wav"

#define LOOP_LEN 192
#define QUANT 3

#define DEFAULT_BPM 140

#define FONT "notepad"
#define FONT_SIZE 12

static pthread_mutex_t render_lock = PTHREAD_MUTEX_INITIALIZER;

typedef void (*Evas_Event_Callback) (void *data, Evas *e, Evas_Object *obj, void *event_info);

static char * progname = "Beatfish";

Evas *evas = NULL;
Evas_Object * cursor;

int win_w = WIDTH;
int win_h = HEIGHT;

RemixEnv * remix;
RemixDeck * drums_deck, * mix_deck;

RemixSquareTone * square1, * square2;
RemixBase * kick_sf, * snare_sf, * congahigh_sf, * congalow_sf,
  * closedhat_sf, * openhat_sf, * clap_sf, * clave_sf;

RemixTrack * mix_track;
RemixLayer * drums_layer, * delay_layer, * vcf303_layer, * monitor_layer;

RemixBase * delay1_fx, * delay2_fx, * vcf303_fx;
RemixMonitor * monitor;

float delay1_x, delay2_x;

RemixBase *
sndfile_reader_new (char * pathname)
{
  static RemixPlugin * sf_plugin = NULL;
  RemixBase * base;
  CDSet * sf_parms;
  int sf_path_key;

  if (sf_plugin == NULL) {
    sf_plugin = remix_find_plugin (remix, "builtin::sndfile_reader");
  }

  if (sf_plugin == NULL) {
    fprintf (stderr, "%s: Could not find soundrender sndfile reader plugin\n",
	     progname);
    exit (1);
  }

  sf_parms = cd_set_new (remix);
  sf_path_key = remix_get_init_parameter_key (remix, sf_plugin, "path");
  sf_parms = cd_set_insert (remix, sf_parms, sf_path_key,
			    CD_STRING(pathname));

  base = remix_new (remix, sf_plugin, sf_parms);

  return base;
}

static RemixPlugin * delay_plugin = NULL;
static RemixPlugin * vcf303_plugin = NULL;

static void
setup_decks (void)
{
  RemixEnvelope * envelope;
  RemixSound * sound;

  remix = remix_init ();
  remix_set_tempo (remix, DEFAULT_BPM);
  remix_set_channels (remix, REMIX_MONO);

  drums_deck = remix_deck_new (remix);

  square1 = remix_squaretone_new (remix, 220);
  square2 = remix_squaretone_new (remix, 238);

  kick_sf = sndfile_reader_new (KICK_PATH);
  snare_sf = sndfile_reader_new (SNARE_PATH);
  congahigh_sf = sndfile_reader_new (CONGAHIGH_PATH);
  congalow_sf = sndfile_reader_new (CONGALOW_PATH);
  closedhat_sf = sndfile_reader_new (CLOSEDHAT_PATH);
  openhat_sf = sndfile_reader_new (OPENHAT_PATH);
  clap_sf = sndfile_reader_new (CLAP_PATH);
  clave_sf = sndfile_reader_new (CLAVE_PATH);

  mix_deck = remix_deck_new (remix);
  mix_track = remix_track_new (remix, mix_deck);
  
  drums_layer = remix_layer_new_ontop (remix, mix_track, REMIX_TIME_BEAT24S);
  remix_sound_new (remix, (RemixBase *)drums_deck, drums_layer,
		  REMIX_BEAT24S(0), REMIX_BEAT24S(LOOP_LEN));

  envelope = remix_envelope_new (remix, REMIX_ENVELOPE_LINEAR);
  remix_envelope_add_point (remix, envelope, REMIX_BEAT24S(0), 0.2);

  delay_plugin = remix_find_plugin (remix, "ladspa::1061");

  if (delay_plugin == NULL) {
    printf ("%s: Could not find Feedback Delay 5s LADSPA plugin (id 1061).\n"
	    "You should install CMT from http://www.ladspa.org/cmt\n",
	    progname);
  } else {

    delay1_fx = remix_new (remix, delay_plugin, CD_EMPTY_SET);
    remix_set_parameter (remix, delay1_fx, 0, CD_FLOAT(0.1)); /* Delay (s) */
    remix_set_parameter (remix, delay1_fx, 1, CD_FLOAT(0.2)); /* Dry/Wet */
    remix_set_parameter (remix, delay1_fx, 2, CD_FLOAT(0.2)); /* Feedback */

    delay_layer = remix_layer_new_ontop (remix, mix_track, REMIX_TIME_BEAT24S);
    sound = remix_sound_new (remix, delay1_fx, delay_layer,
			    REMIX_BEAT24S(0), REMIX_BEAT24S(LOOP_LEN));
    /*  remix_sound_set_blend_envelope (remix, sound, envelope);*/
  
    delay2_fx = remix_new (remix, delay_plugin, CD_EMPTY_SET);
    remix_set_parameter (remix, delay2_fx, 0, CD_FLOAT(0.2)); /* Delay (s) */
    remix_set_parameter (remix, delay2_fx, 1, CD_FLOAT(0.1)); /* Dry/Wet */
    remix_set_parameter (remix, delay2_fx, 2, CD_FLOAT(0.1)); /* Feedback */

    delay_layer = remix_layer_new_ontop (remix, mix_track, REMIX_TIME_BEAT24S);
    sound = remix_sound_new (remix, delay2_fx, delay_layer,
			    REMIX_BEAT24S(0), REMIX_BEAT24S(LOOP_LEN));
    /*remix_sound_set_blend_envelope (remix, sound, envelope);*/
  }
    
  vcf303_plugin = remix_find_plugin (remix, "ladspa::1224");

  if (vcf303_plugin == NULL) {
    printf ("%s: Could not find VCF 303 LADSPA plugin (id 1224).\n"
	    "You should install CMT from http://www.ladspa.org/cmt\n",
	    progname);
  } else {
    vcf303_fx = remix_new (remix, vcf303_plugin, CD_EMPTY_SET);
    remix_set_parameter (remix, vcf303_fx, 0, CD_FLOAT(1)); /* Trigger */
    remix_set_parameter (remix, vcf303_fx, 1, CD_FLOAT(0.6)); /* Cutoff */
    remix_set_parameter (remix, vcf303_fx, 2, CD_FLOAT(0.7)); /* Resonance */
    remix_set_parameter (remix, vcf303_fx, 3, CD_FLOAT(0.4)); /* Env. Mod. */
    remix_set_parameter (remix, vcf303_fx, 4, CD_FLOAT(0.3)); /* Decay */
  
    vcf303_layer = remix_layer_new_ontop (remix, mix_track, REMIX_TIME_BEAT24S);
    remix_sound_new (remix, vcf303_fx, vcf303_layer,
		    REMIX_BEAT24S(0), REMIX_BEAT24S(LOOP_LEN));
  }

#if 0
  monitor = remix_monitor_new (remix);
  monitor_layer = remix_layer_new_ontop (remix, mix_track, REMIX_TIME_BEAT24S);
  remix_sound_new (remix, (RemixBase *)monitor, monitor_layer,
		  REMIX_BEAT24S(0), REMIX_BEAT24S(LOOP_LEN));
#endif
}

#if 0
void
zoom_in (void)
{
  double x, y, w, h;
  int sx, sy, sw, sh;

  evas_output_viewport_get (evas, &x, &y, &w, &h);

  x += w/4;
  w /= 2;

  evas_output_viewport_set (evas, x, y, w, h);

  sx = evas_coord_world_x_to_screen (evas, x);
  sy = evas_coord_world_y_to_screen (evas, y);
  sw = evas_coord_world_x_to_screen (evas, w);
  sh = evas_coord_world_y_to_screen (evas, h);

  evas_damage_rectangle_add (evas, sx, sy, sw, sh);
}

void
zoom_out (void)
{
  double x, y, w, h;
  int sx, sy, sw, sh;

  evas_output_viewport_get (evas, &x, &y, &w, &h);
  x -= w/2;
  w *= 2;

  if (w >= WIDTH) {
    x = 0;
    w = WIDTH;
  }

  evas_output_viewport_set (evas, x, y, w, h);

  sx = evas_coord_world_x_to_screen (evas, x);
  sy = evas_coord_world_y_to_screen (evas, y);
  sw = evas_coord_world_x_to_screen (evas, w);
  sh = evas_coord_world_y_to_screen (evas, h);

  evas_damage_rectangle_add (evas, sx, sy, sw, sh);
}
#endif

static void
disable_rect (double x, double y, double w, double h)
{
  Evas_Object * ob;

  ob = evas_object_rectangle_add (evas);
  evas_object_move (ob, x, y);
  evas_object_resize (ob, w, h);
  evas_object_color_set (ob, 200, 200, 200, 160);
  evas_object_layer_set (ob, 4);
  evas_object_show (ob);
}

void
delay_reset (RemixBase * delay_fx, float canvas_x)
{
 int beat;
 RemixTime ts;
 float value;
 
 beat = LOOP_LEN * canvas_x / WIDTH;
 ts = remix_time_convert (remix, REMIX_BEAT24S(beat), REMIX_TIME_BEAT24S,
		       REMIX_TIME_SECONDS);
 remix_set_parameter (remix, delay_fx, 0, CD_FLOAT(ts.seconds)); 
}

static Evas_Object * tempo_label;

void
tempo_set_x (Evas_Object * slider, float canvas_x)
{
  RemixTempo new_tempo;
  static char buf[32];

  if (canvas_x > 0 && canvas_x < WIDTH/2) {
    new_tempo =  60 + 180 * (canvas_x / (WIDTH/2));
    snprintf (buf, 32, "Tempo: %.1f BPM", new_tempo);
    
    remix_set_tempo (remix, new_tempo);
    
    evas_object_text_text_set (tempo_label, buf);
    evas_object_resize (slider, canvas_x, HEIGHT/16 - 16);

    delay_reset (delay1_fx, delay1_x);
    delay_reset (delay2_fx, delay2_x);
  }
}

void
tempo_mouse_down(void *data, Evas * e, Evas_Object * obj,
		  Evas_Event_Mouse_Down * ev)
{
  Evas_Object * slider = (Evas_Object *)data;

  evas_object_data_set (obj, "moving", (void *)1);
  tempo_set_x (slider, ev->canvas.x);
}

void
tempo_mouse_move (void *data, Evas * e, Evas_Object * obj,
		  Evas_Event_Mouse_Move * ev)
{
  Evas_Object * slider = (Evas_Object *)data;

  if (evas_object_data_get (obj, "moving") != NULL) {
    tempo_set_x (slider, ev->cur.canvas.x);
  }
}

void
tempo_mouse_up (void *data, Evas * e, Evas_Object * obj,
		 Evas_Event_Mouse_Move * ev)
{
  evas_object_data_set (obj, "moving", NULL);
}

static void
setup_tempo_slider (void)
{
  double track_y;
  Evas_Object * bg, * ob;
  CDScalar value;
  char buf[32];
  RemixTempo tempo = remix_get_tempo (remix);

  track_y = 3 * HEIGHT/16;

  ob = evas_object_image_add (evas);
  evas_object_image_file_set(ob, IM "track_bg.png", NULL);
  evas_object_move(ob, 0, track_y);
  evas_object_resize(ob, WIDTH/2, HEIGHT/16);
  evas_object_image_fill_set(ob, 0, 0, WIDTH/2, HEIGHT/16);
  evas_object_color_set (ob, 240, 240, 240, 255);
  evas_object_layer_set(ob, 0);
  evas_object_show(ob);
  
  bg = ob;

  ob = evas_object_rectangle_add (evas);
  evas_object_move (ob, 0, track_y + 8);
  evas_object_resize (ob, ((tempo - 60) / 180) * WIDTH/2, HEIGHT/16 - 16);
  evas_object_color_set (ob, 250, 250, 120, 100);
  evas_object_pass_events_set (ob, 1);
  evas_object_show (ob);

  evas_object_event_callback_add(bg, EVAS_CALLBACK_MOUSE_DOWN,
				 (Evas_Event_Callback)tempo_mouse_down, ob);
  evas_object_event_callback_add(bg, EVAS_CALLBACK_MOUSE_UP,
				 (Evas_Event_Callback)tempo_mouse_up, ob);
  evas_object_event_callback_add(bg, EVAS_CALLBACK_MOUSE_MOVE,
				 (Evas_Event_Callback)tempo_mouse_move, ob);

  snprintf (buf, 32, "Tempo: %.1f BPM", tempo);

  ob = evas_object_text_add (evas);
  evas_object_text_font_set (ob, FONT, FONT_SIZE);
  evas_object_text_text_set (ob, buf);
  evas_object_move (ob, 10, track_y+8);
  evas_object_layer_set (ob, 1);
  evas_object_color_set (ob, 0, 0, 0, 255);
  evas_object_pass_events_set (ob, 1);
  evas_object_show (ob);

  tempo_label = ob;
}

/* VCF 303 */

typedef struct _MD_VCF303 MD_VCF303;

struct _MD_VCF303 {
  RemixBase * vcf303_fx;
  int param_key;
  Evas_Object * slider;
};

void
vcf303_set_x (MD_VCF303 * md, float canvas_x)
{
  int beat;
  RemixTime ts;
  float value;

  remix_set_parameter (remix, md->vcf303_fx, md->param_key,
		      CD_FLOAT((canvas_x - WIDTH/2) / (WIDTH/2)));
  evas_object_resize (md->slider, canvas_x - WIDTH/2, HEIGHT/16 - 16);
}

void
vcf303_mouse_down(void *data, Evas * e, Evas_Object * obj,
		  Evas_Event_Mouse_Down * ev)
{
  MD_VCF303 * md = (MD_VCF303 *)data;

  evas_object_data_set (obj, "moving", (void *)1);

  vcf303_set_x (md, ev->canvas.x);
}

void
vcf303_mouse_move (void *data, Evas * e, Evas_Object * obj,
		  Evas_Event_Mouse_Move * ev)
{
  MD_VCF303 * md = (MD_VCF303 *)data;

  if (evas_object_data_get (obj, "moving") != NULL) {
    vcf303_set_x (md, ev->cur.canvas.x);
  }
}

void
vcf303_mouse_up (void *data, Evas * e, Evas_Object * obj,
		 Evas_Event_Mouse_Move * ev)
{
  evas_object_data_set (obj, "moving", NULL);
}

static void
setup_vcf303_slider (int index, char * name)
{
  MD_VCF303 * md;
  double track_y;
  Evas_Object * ob;
  CDScalar value;

  md = malloc (sizeof(MD_VCF303));
  md->vcf303_fx = vcf303_fx;
  md->param_key = index;

  track_y = (index-1) * HEIGHT/16;

  ob = evas_object_image_add (evas);
  evas_object_image_file_set(ob, IM "track_bg.png", NULL);
  evas_object_move(ob, WIDTH/2, track_y);
  evas_object_resize(ob, WIDTH/2, HEIGHT/16);
  evas_object_image_fill_set(ob, 0, 0, WIDTH/2, HEIGHT/16);
  evas_object_color_set (ob, 250, 240, 220, 255);
  evas_object_layer_set(ob, 0);
  evas_object_show(ob);
  evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_DOWN,
				 (Evas_Event_Callback)vcf303_mouse_down, md);
  evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_UP,
				 (Evas_Event_Callback)vcf303_mouse_up, md);
  evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_MOVE,
				 (Evas_Event_Callback)vcf303_mouse_move, md);

  value = remix_get_parameter (remix, vcf303_fx, index);

  if (vcf303_plugin != NULL) {
    ob = evas_object_rectangle_add (evas);
    evas_object_move (ob, WIDTH/2, track_y + 8);
    evas_object_resize (ob, value.s_float * WIDTH/2, HEIGHT/16 - 16);
    evas_object_color_set (ob, 250, 250, 120, 100);
    evas_object_pass_events_set (ob, 1);
    evas_object_show (ob);
    md->slider = ob;
  }

  ob = evas_object_text_add (evas);
  evas_object_text_font_set (ob, FONT, FONT_SIZE);
  evas_object_text_text_set (ob, name);
  evas_object_move (ob, WIDTH/2 + 10, track_y+8);
  evas_object_layer_set (ob, 1);
  evas_object_color_set (ob, 0, 0, 0, 255);
  evas_object_pass_events_set (ob, 1);
  evas_object_show (ob);

}

static void
setup_vcf303 (void)
{
  Evas_Object * ob;

  /* These names can be read from the plugin but this code is DRUNK */
  setup_vcf303_slider (1, "Cutoff");
  setup_vcf303_slider (2, "Resonance");
  setup_vcf303_slider (3, "Envelope Modulation");
  setup_vcf303_slider (4, "Decay");

  ob = evas_object_text_add (evas);
  evas_object_text_font_set (ob, FONT, FONT_SIZE*4.4);
  evas_object_text_text_set (ob, "VCF303");
  evas_object_move (ob, WIDTH/2 + 30, 30);
  evas_object_layer_set (ob, 1);
  evas_object_color_set (ob, 200, 200, 200, 100);
  evas_object_pass_events_set (ob, 1);
  evas_object_show (ob);

  if (vcf303_plugin == NULL) {
    disable_rect (WIDTH/2, 0, WIDTH/2, HEIGHT/4);

    ob = evas_object_text_add (evas);
    evas_object_text_font_set (ob, FONT, FONT_SIZE*2);
    evas_object_text_text_set (ob, "Plugin not found");
    evas_object_move (ob, WIDTH/2 + 40, HEIGHT/8 - 20);
    evas_object_layer_set (ob, 5);
    evas_object_color_set (ob, 200, 20, 20, 200);
    evas_object_pass_events_set (ob, 1);
    evas_object_show (ob);

  }
}

/* Scrub */

void
scrub_set_x (float canvas_x)
{
  RemixCount length, offset, n = 0;

  pthread_mutex_lock (&render_lock);

  length = remix_length (remix, mix_deck);

  remix_seek (remix, mix_deck, (RemixCount)(length * canvas_x / WIDTH), SEEK_SET);

  pthread_mutex_unlock (&render_lock);
}

void
scrub_mouse_down(void *data, Evas * e, Evas_Object * obj,
		 Evas_Event_Mouse_Down * ev)
{
  evas_object_data_set (obj, "moving", (void *)1);

  scrub_set_x (ev->canvas.x);
}

void
scrub_mouse_move (void *data, Evas * e, Evas_Object * obj,
		  Evas_Event_Mouse_Move * ev)
{
  if (evas_object_data_get (obj, "moving") != NULL) {
    scrub_set_x (ev->cur.canvas.x);
  }
}

void
scrub_mouse_up (void *data, Evas * e, Evas_Object * obj,
		Evas_Event_Mouse_Move * ev)
{
  evas_object_data_set (obj, "moving", NULL);
}

static void
setup_scrub (void)
{
  double track_y;
  Evas_Object * ob;

  track_y = HEIGHT/4;

  ob = evas_object_image_add (evas);
  evas_object_image_file_set(ob, IM "track_bg.png", NULL);
  evas_object_move(ob, 0, track_y);
  evas_object_resize(ob, WIDTH, (HEIGHT/2)/4);
  evas_object_image_fill_set(ob, 0, 0, WIDTH/4, (HEIGHT/2)/8);
  evas_object_color_set (ob, 220, 250, 220, 255);
  evas_object_layer_set(ob, 0);
  evas_object_show(ob);
  evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_DOWN,
				 (Evas_Event_Callback)scrub_mouse_down, NULL);
  evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_UP,
				 (Evas_Event_Callback)scrub_mouse_up, NULL);
  evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_MOVE,
				 (Evas_Event_Callback)scrub_mouse_move, NULL);

  ob = evas_object_text_add (evas);
  evas_object_text_font_set (ob, FONT, FONT_SIZE*1.4);
  evas_object_text_text_set (ob, "Scrub");
  evas_object_move (ob, 10, track_y+12);
  evas_object_layer_set (ob, 1);
  evas_object_color_set (ob, 100, 100, 100, 200);
  evas_object_pass_events_set (ob, 1);
  evas_object_show (ob);

}

/* Delay */

typedef struct _MD_Delay MD_Delay;

struct _MD_Delay {
  RemixBase * delay_fx;
  Evas_Object * slider;
  float * delay_x;
};

void
delay_set_x (MD_Delay * md, float canvas_x)
{
  *md->delay_x = canvas_x;
  delay_reset (md->delay_fx, canvas_x);
  evas_object_resize (md->slider, canvas_x, HEIGHT/16 - 16);  
}

void
delay_mouse_down(void *data, Evas * e, Evas_Object * obj,
		 Evas_Event_Mouse_Down * ev)
{
  MD_Delay * md = (MD_Delay *)data;

#if 0
  if (ev->button == 4) {
    zoom_in ();
  } else if (ev->button == 5) {
    zoom_out ();
  } else {
    evas_object_data_set (obj, "moving", (void *)1);
    
    delay_set_x (md, ev->canvas.x);
  }
#else
  evas_object_data_set (obj, "moving", (void *)1);
  delay_set_x (md, ev->canvas.x);
#endif
}

void
delay_mouse_move (void *data, Evas * e, Evas_Object * obj,
		  Evas_Event_Mouse_Move * ev)
{
  MD_Delay * md = (MD_Delay *)data;

  if (evas_object_data_get (obj, "moving") != NULL) {
    delay_set_x (md, ev->cur.canvas.x);
  }
}

void
delay_mouse_up (void *data, Evas * e, Evas_Object * obj,
		Evas_Event_Mouse_Move * ev)
{
  evas_object_data_set (obj, "moving", NULL);
}


static void
setup_delay (int index, RemixBase * delay_fx, float * delay_x)
{
  MD_Delay * md;
  RemixTrack * track;
  RemixLayer * layer;
  Evas_Object * ob;
  int iw, ih;
  double track_y;
  CDScalar delay_seconds;
  RemixTime loop_seconds;

  track_y = 3*HEIGHT/8 + index * HEIGHT / (2 * 8);

  md = malloc (sizeof(MD_Delay));

  md->delay_fx = delay_fx;
  md->delay_x = delay_x;
  
  ob = evas_object_image_add (evas);
  evas_object_image_file_set(ob, IM "track_bg.png", NULL);
  evas_object_move(ob, 0, track_y);
  evas_object_resize(ob, WIDTH, (HEIGHT/2)/8);
  evas_object_image_fill_set(ob, 0, 0, WIDTH/8, (HEIGHT/2)/8);
  evas_object_color_set (ob, 250, 240, 220, 255);
  evas_object_layer_set(ob, 0);
  evas_object_show(ob);

  if (delay_plugin != NULL) {
    evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_DOWN,
				   (Evas_Event_Callback)delay_mouse_down, md);
    evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_UP,
				   (Evas_Event_Callback)delay_mouse_up, md);
    evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_MOVE,
				   (Evas_Event_Callback)delay_mouse_move, md);
    
    
    ob = evas_object_rectangle_add (evas);
    evas_object_move (ob, 0, track_y + 8);
    evas_object_color_set (ob, 250, 250, 120, 100);
    evas_object_pass_events_set (ob, 1);
    evas_object_show (ob);
    md->slider = ob;

    delay_seconds = remix_get_parameter (remix, delay_fx, 0);
    loop_seconds = remix_time_convert (remix, REMIX_BEAT24S(LOOP_LEN),
				      REMIX_TIME_BEAT24S, REMIX_TIME_SECONDS);
    *delay_x = delay_seconds.s_float / loop_seconds.seconds * WIDTH;
    evas_object_resize (ob, *delay_x, HEIGHT/16 - 16);
  }

  ob = evas_object_text_add (evas);
  evas_object_text_font_set (ob, FONT, FONT_SIZE);
  evas_object_text_text_set (ob, "Delay");
  evas_object_move (ob, 10, track_y+8);
  evas_object_layer_set (ob, 1);
  evas_object_color_set (ob, 0, 0, 0, 255);
  evas_object_pass_events_set (ob, 1);
  evas_object_show (ob);

  if (delay_plugin == NULL) {
    disable_rect (0, track_y, WIDTH, HEIGHT/16);

    ob = evas_object_text_add (evas);
    evas_object_text_font_set (ob, FONT, FONT_SIZE);
    evas_object_text_text_set (ob, "Plugin not found");
    evas_object_move (ob, 180, track_y+8);
    evas_object_layer_set (ob, 5);
    evas_object_color_set (ob, 200, 20, 20, 200);
    evas_object_pass_events_set (ob, 1);
    evas_object_show (ob);

  }

}

/* Track GUI */

typedef struct _MDTrack MDTrack;

struct _MDTrack {
  RemixBase * sf_base;
  RemixLayer * layer;
  double track_y;
  char * img_path;
};

void
sound_mouse_down(void *data, Evas * e, Evas_Object * obj,
		 Evas_Event_Mouse_Down * ev)
{
  RemixBase * sound;

  if (ev->button == 1) {
    evas_object_data_set (obj, "moving", (void *)1);
  } else if (ev->button == 3) {
    sound = evas_object_data_get (obj, "sound");

    pthread_mutex_lock (&render_lock);
    remix_destroy (remix, sound);
    pthread_mutex_unlock (&render_lock);
    
    evas_object_del (obj);
  }
}

void
sound_mouse_up(void *data, Evas * e, Evas_Object * obj,
	       Evas_Event_Mouse_Up * ev)
{
  double ox, oy;
  int beat;

  evas_object_geometry_get (obj, &ox, &oy, NULL, NULL);
  beat = (int) ((LOOP_LEN/QUANT) * ox / WIDTH) * QUANT;
  evas_object_move (obj, WIDTH * beat / LOOP_LEN, oy);

  evas_object_data_set (obj, "moving", NULL);
}

void
sound_mouse_move(void *data, Evas * e, Evas_Object * obj,
		 Evas_Event_Mouse_Move * ev)
{
  RemixBase * sound;
  double ox, oy, nx;
  int beat;
  
  if (evas_object_data_get (obj, "moving") != NULL) {
    evas_object_geometry_get (obj, &ox, &oy, NULL, NULL);
    sound = evas_object_data_get (obj, "sound");

    beat = (int) ((LOOP_LEN/QUANT) * ev->cur.canvas.x / WIDTH) * QUANT;

    pthread_mutex_lock (&render_lock);
    remix_sound_move (remix, sound, REMIX_BEAT24S(beat));
    pthread_mutex_unlock (&render_lock);

    evas_object_move (obj, WIDTH * beat / LOOP_LEN, oy);
  }
}

void
track_mouse_down(void *data, Evas * e, Evas_Object * obj,
		 Evas_Event_Mouse_Down * ev)
{
  MDTrack * md = (MDTrack *)data;
  Evas_Object * ob;
  int iw, ih;
  RemixBase * sound;
  int beat;

  if (ev->button == 1) {
    ob = evas_object_image_add(evas);
    evas_object_image_file_set(ob, md->img_path, NULL);
    evas_object_image_size_get(ob, &iw, &ih);
    evas_object_move(ob, ev->canvas.x - iw/2, md->track_y + 2);
    evas_object_resize(ob, (int)(iw/2), HEIGHT/16-4);
    evas_object_image_fill_set(ob, 0, 0, (int)(iw/2), HEIGHT/16-4);
    evas_object_color_set (ob, 255, 255, 255, 140);
    evas_object_layer_set(ob, 1);
    evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_DOWN,
				   (Evas_Event_Callback)sound_mouse_down,
				   NULL);
    evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_UP,
				   (Evas_Event_Callback)sound_mouse_up,
				   NULL);
    evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_MOVE,
				   (Evas_Event_Callback)sound_mouse_move,
				   NULL);

    beat = (int) ((LOOP_LEN/QUANT) * ev->canvas.x / WIDTH) * QUANT;

    pthread_mutex_lock (&render_lock);
    sound = remix_sound_new (remix, md->sf_base, md->layer,
			     REMIX_BEAT24S(beat), REMIX_BEAT24S(12));
    pthread_mutex_unlock (&render_lock);

    evas_object_move (ob, WIDTH * beat / LOOP_LEN, md->track_y + 2);
    evas_object_show(ob);

    evas_object_data_set (ob, "sound", sound);
  }
}

static void
setup_track (int index, char * name, RemixBase * sf_base, RemixPCM gain,
	     char * img_path)
{
  MDTrack * md;
  RemixTrack * track;
  RemixLayer * layer;
  Evas_Object * ob;
  int iw, ih;
  double track_y;

  track = remix_track_new (remix, drums_deck);
  remix_track_set_gain (remix, track, gain);

  layer = remix_layer_new_ontop (remix, track, REMIX_TIME_BEAT24S);

  track_y = HEIGHT/2 + index * HEIGHT / (2 * 8);

  if (sf_base != NULL) {
    md = malloc (sizeof(MDTrack));
    md->sf_base = sf_base;
    md->layer = layer;
    md->track_y = track_y;
    md->img_path = img_path;
  }

  ob = evas_object_image_add (evas);
  evas_object_image_file_set(ob, IM "track_bg.png", NULL);
  evas_object_move(ob, 0, track_y);
  evas_object_resize(ob, WIDTH, (HEIGHT/2)/8);
  evas_object_image_fill_set(ob, 0, 0, WIDTH/8, (HEIGHT/2)/8);
  evas_object_layer_set(ob, 0);
  if (index % 2) {
    evas_object_color_set (ob, 240, 240, 240, 255);
  } else {
    evas_object_color_set (ob, 220, 220, 220, 255);
  }
  evas_object_show(ob);

  if (sf_base != NULL) {
    evas_object_event_callback_add(ob, EVAS_CALLBACK_MOUSE_DOWN,
				   (Evas_Event_Callback)track_mouse_down, md);
  }

  ob = evas_object_text_add (evas);
  evas_object_text_font_set (ob, FONT, FONT_SIZE);
  evas_object_text_text_set (ob, name);
  evas_object_move (ob, 10, track_y+8);
  evas_object_layer_set (ob, 1);
  evas_object_color_set (ob, 60, 60, 60, 200);
  evas_object_pass_events_set (ob, 1);
  evas_object_show (ob);

  if (sf_base == NULL) {
    disable_rect (0, track_y, WIDTH, HEIGHT/16);

    ob = evas_object_text_add (evas);
    evas_object_text_font_set (ob, FONT, FONT_SIZE);
    evas_object_text_text_set (ob, "Sample not found");
    evas_object_move (ob, 180, track_y+8);
    evas_object_layer_set (ob, 5);
    evas_object_color_set (ob, 200, 20, 20, 200);
    evas_object_pass_events_set (ob, 1);
    evas_object_show (ob);

  }
}

void
setup_gui (void)
{
   Evas_Object        *ob;
   int                 iw, ih;

   evas_font_path_prepend(evas, PACKAGE_DATA_DIR);
#if 1
   evas_image_cache_set(evas, 1024 * 1024);
   evas_font_cache_set(evas, 256 * 1024);
#else
   evas_image_cache_set(evas, 0);
   evas_font_cache_set(evas, 0);
#endif

   ob = evas_object_rectangle_add(evas);
   evas_object_move(ob, 0, 0);
   evas_object_resize(ob, win_w, win_h);
   evas_object_color_set(ob, 0, 0, 0, 255);
   evas_object_show(ob);

   ob = evas_object_image_add (evas);
   evas_object_image_file_set(ob, IM "track_bg.png", NULL);
   evas_object_move(ob, 0, 0);
   evas_object_resize(ob, WIDTH/2, 3*HEIGHT/16);
   evas_object_image_fill_set(ob, 0, 0, WIDTH/2, HEIGHT/8);
   evas_object_layer_set(ob, 0);
   evas_object_color_set (ob, 240, 240, 240, 255);
   evas_object_show(ob);

   ob = evas_object_text_add (evas);
   evas_object_text_font_set (ob, FONT, FONT_SIZE * 3.6);
   evas_object_text_text_set (ob, progname);
   evas_object_move (ob, 8, 12);
   evas_object_layer_set (ob, 1);
   evas_object_color_set (ob, 60, 60, 60, 200);
   evas_object_pass_events_set (ob, 1);
   evas_object_show (ob);

   ob = evas_object_text_add (evas);
   evas_object_text_font_set (ob, FONT, FONT_SIZE);
   evas_object_text_text_set (ob, "Banana code by Conrad Parker 2003");
   evas_object_move (ob, 12, 5 * HEIGHT/32);
   evas_object_layer_set (ob, 1);
   evas_object_color_set (ob, 60, 60, 60, 200);
   evas_object_pass_events_set (ob, 1);
   evas_object_show (ob);

   setup_tempo_slider ();

   setup_vcf303 ();

   setup_scrub ();

   setup_delay (0, delay1_fx, &delay1_x);
   setup_delay (1, delay2_fx, &delay2_x);

#if 0
   setup_track (0, "Clave", clave_sf, 1.0, IM "drum.png");
   setup_track (1, "Clap", clap_sf, 1.0, IM "drum.png");
#else
   setup_track (0, "Square B", square2, 0.1, IM "note.png");
   setup_track (1, "Square A", square1, 0.1, IM "note.png");
#endif
   setup_track (2, "Open Hat", openhat_sf, 1.0, IM "drum.png");
   setup_track (3, "Closed Hat", closedhat_sf, 1.0, IM "drum.png");
   setup_track (4, "Conga High", congahigh_sf, 1.0, IM "drum.png");
   setup_track (5, "Conga Low", congalow_sf, 1.0, IM "drum.png");
   setup_track (6, "Snare", snare_sf, 1.0, IM "drum.png");
   setup_track (7, "Kick", kick_sf, 1.0, IM "drum.png");

#ifdef TINY_CURSOR
   cursor = evas_object_rectangle_add (evas);
   evas_object_resize (cursor, 2, HEIGHT/8);
#else
   cursor = evas_object_image_add (evas);
   evas_object_image_file_set (cursor, IM "cursor.png", NULL);
   evas_object_image_fill_set (cursor, 0, 0, 32, 3*win_h/4);
   evas_object_resize (cursor, 32, 3*win_h/4);
#endif
   evas_object_move (cursor, 0, win_h/4);
   evas_object_color_set (cursor, 200, 240, 200, 255);
   evas_object_pass_events_set (cursor, 1);
   evas_object_show (cursor);
}

#ifdef USE_JACK
jack_port_t * output_port;
RemixStream * output_stream = NULL;
static jack_nframes_t last_nframes = 0;
static RemixPCM * last_pcm = NULL;
static RemixPCM * pcm = NULL;

int
beatfish_process (jack_nframes_t nframes, void * arg)
{
  RemixCount length, offset;
  
  pcm = (RemixPCM *) jack_port_get_buffer (output_port, nframes);

  pthread_mutex_lock (&render_lock);

  length = remix_length (remix, mix_deck);
  offset = remix_tell (remix, mix_deck);

  if (offset > length) {
    remix_seek (remix, mix_deck, 0, SEEK_SET);
  }

  if ((last_nframes != nframes) || (last_pcm != pcm)) {
    output_stream = remix_stream_new_from_buffers (remix, nframes, &pcm);
    last_nframes = nframes;
    last_pcm = pcm;
  } else {
    remix_seek (remix, output_stream, 0, SEEK_SET);
  }

  remix_process (remix, mix_deck, nframes, RemixNone, output_stream);

  pthread_mutex_unlock (&render_lock);

  return 0;
}

void
beatfish_shutdown (void * arg)
{
  printf ("Beatfish JACK client shut down!\n");
  exit (1);
}

void
setup_jack (char * client_name)
{
  jack_client_t * client;
  const char ** ports;

  if ((client = jack_client_new (client_name)) == 0) {
    exit (1);
  }

  jack_set_process_callback (client, beatfish_process, 0);

  jack_on_shutdown (client, beatfish_shutdown, 0);

  remix_set_samplerate (remix, jack_get_sample_rate(client));

  output_port = jack_port_register (client, "output",
				    JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput,
				    0);

  jack_activate (client);

  if ((ports = jack_get_ports (client, NULL, NULL,
			       JackPortIsPhysical|JackPortIsInput)) == NULL) {
    fprintf (stderr, "Cannot find any physical playback ports\n");
    exit (1);
  }

  if (jack_connect (client, jack_port_name (output_port), ports[0])) {
    fprintf (stderr, "Cannot connect output ports\n");
  }

  free (ports);
}
#endif

void
setup (void)
{
  char client_name[32];

  snprintf (client_name, 32, "beatfish_%d", getpid());

  setup_decks ();
  setup_gui ();

#ifdef USE_JACK
  setup_jack (client_name);
#endif
}

void
update_cursor (void)
{
  RemixCount length, offset, n = 0;
  static int i = 0;

  pthread_mutex_lock (&render_lock);
  length = remix_length (remix, mix_deck);
  offset = remix_tell (remix, mix_deck);
  pthread_mutex_unlock (&render_lock);

#if 1
  if (offset > length) {
    n = 0;
  } else {
    n = MIN (1024, length - offset);
#ifndef USE_JACK
    n = remix_process (remix, mix_deck, n, RemixNone, RemixNone);
#endif
    if (n == -1) {
      printf ("remix_process error\n");
    }
  }

  if (n == 0) {
    remix_seek (remix, mix_deck, 0, SEEK_SET);
    offset = 0;
  }
#endif

#ifdef TINY_CURSOR
  evas_object_move (cursor, (win_w * offset/length) - 2, win_h/4);
#else
  evas_object_move (cursor, (win_w * offset/length) - 32, win_h/4);
#endif
}
